#include <plib.h>
#include <stdio.h>
#include <xc.h>

#include "hmac.h"

#pragma config FNOSC    = PRIPLL
#pragma config POSCMOD  = EC       // IC2 oscillator
#pragma config FPLLIDIV = DIV_2    // PLL Input Divider
#pragma config FPLLMUL  = MUL_20   // PLL Multiplier
#pragma config FPLLODIV = DIV_1    // PLL Output Divider
#pragma config FPBDIV   = DIV_8    // PBCLK Divider (for timer)

#pragma config PWP = OFF // allow program flash write
#pragma config BWP = OFF // allow program flash write

__attribute__((aligned(4096))) const unsigned char settings[4096] =
   {0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF};

// settings:
// secret (8 bytes)
// counter (4 bytes)

void UARTInit() {
  // disable interrupts from UART1
  ConfigIntUART1(UART_ERR_INT_DIS | UART_RX_INT_DIS | UART_TX_INT_DIS);
  OpenUART1(UART_EN | UART_NO_PAR_8BIT | UART_1STOPBIT,
            UART_TX_ENABLE | UART_RX_ENABLE,
            64); // 9600 baud @ PBCLK = 10MHz
}

void LEDInit() {
  // LED configuration
  PORTSetPinsDigitalOut (IOPORT_G, BIT_12|BIT_13| BIT_14|BIT_15);
  PORTClearBits(IOPORT_G, BIT_12|BIT_13| BIT_14|BIT_15);
}

void LEDToggle() {
  PORTToggleBits(IOPORT_G, BIT_12|BIT_13| BIT_14|BIT_15);
}

int isHexChar(char c) {
  if ((c >= 'A') && (c <= 'F')) return 1;
  if ((c >= 'a') && (c <= 'f')) return 1;
  if ((c >= '0') && (c <= '9')) return 1;
  return 0;
}

unsigned char getHexChar() {
  unsigned char c;
  do {
     while (DataRdyUART1() == 0) ;
     c = ReadUART1();
  } while (! isHexChar(c));
  return c;
}

void getBytes(unsigned char *bytes, unsigned len) {
  unsigned int i;
  char buf[3];
  unsigned int v;
  for (i=0; i<len; i++) {
    buf[0] = getHexChar();
    buf[1] = getHexChar();
    buf[2] = 0;
    sscanf(buf,"%x",&v);
    bytes[i] = v & 0xFF;
  }
}

void putBytes(unsigned char *bytes, unsigned len) {
  unsigned int i;
  char buf[3];
  for (i=0; i<len; i++) {
    sprintf(buf,"%02x", bytes[i]);
    while (BusyUART1());
    WriteUART1(buf[0]);
    while (BusyUART1());
    WriteUART1(buf[1]);
  }
}

void scanFor(unsigned char f) {
  unsigned char c;
  do {
     while (DataRdyUART1() == 0) ;
     c = ReadUART1();
  } while (c != f);
}

void getID(unsigned char id[4]) {
  scanFor('[');
  getBytes(id, 4);
  scanFor(']');
}

void putChallenge(unsigned char challenge[20]) {
  while (BusyUART1());
  WriteUART1('[');
  putBytes(challenge, 20);
  while (BusyUART1());
  WriteUART1(']');
}

void getResponse(unsigned char response[20]) {
  scanFor('[');
  getBytes(response, 20);
  scanFor(']');
}

int correctResponse(unsigned char response[20],
                    unsigned char challenge[20]) {
  int correct = 1;
  unsigned i;
  for (i=0; i<20; i++) {
     correct = correct & (response[i] == challenge[i]);
  }
  return correct;
}

void IncCounter() {
  int *cp, v;
  unsigned int buf[3];
  memcpy(buf, settings, 12);
  buf[2] = buf[2] + 1;
  NVMErasePage((void *)   settings);  
  NVMWriteWord((void *)   settings,     buf[0]);
  NVMWriteWord((void *) &(settings[4]), buf[1]);
  NVMWriteWord((void *) &(settings[8]), buf[2]);
}

int main() {
  unsigned char id[4];         // id of prover
  unsigned char challenge[20]; // challenge to prover
  unsigned char expect[20];    // expected response
  unsigned char response[20];  // expected response

  UARTInit();
  LEDInit();
  LEDToggle();

  while (1) {

      putsUART1("Authentication Ready. Enter ID in format [XXXXXXXX]:\n");
      getID(id);

      sha1_init();
      sha1_update(&(settings[8]), 4);
      sha1_final(challenge);

      hmac(settings, challenge, id, expect);

      IncCounter();
      
      putsUART1("Challenge: ");
      putChallenge(challenge);
      WriteUART1('\n');

      putsUART1("Expected Response: ");
      putChallenge(expect);
      WriteUART1('\n');

      putsUART1("Return Response in format [XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX]:\n");

      getResponse(response);
      if (correctResponse(response, expect)) {
          putsUART1("Authenticated\n");
          LEDToggle();
      }

  }

  return 0;
}